from __future__ import annotations

import os
from typing import Dict, Iterator, List, Literal, Tuple

import numpy as np
import polars as pl
from loguru import logger
from torch.utils.data import Sampler

from alphafold3_pytorch.tensor_typing import typecheck
from alphafold3_pytorch.utils.utils import exists

# constants

CLUSTERING_RESIDUE_MOLECULE_TYPE = Literal["protein", "rna", "dna", "ligand", "peptide"]

# helper functions


def get_chain_count(molecule_type: CLUSTERING_RESIDUE_MOLECULE_TYPE) -> Tuple[int, int, int]:
    """
    Returns the number of protein (or `peptide`), nucleic acid (i.e., `rna` or `dna`), and
    ligand chains in a molecule based on its type.

    Example:
        n_prot, n_nuc, n_ligand = get_chain_count("protein")
    """
    match molecule_type:
        case "protein":
            return 1, 0, 0
        case "rna":
            return 0, 1, 0
        case "dna":
            return 0, 1, 0
        case "ligand":
            return 0, 0, 1
        case "peptide":
            return 1, 0, 0
        case _:
            raise ValueError(f"Unknown molecule type: {molecule_type}")


def calculate_weight(
    alphas: Dict[str, float],
    beta: float,
    n_prot: int,
    n_nuc: int,
    n_ligand: int,
    cluster_size: int,
) -> float:
    """
    Calculates the weight of a chain or an interface according to the formula
    provided in Section 2.5.1 of the AlphaFold 3 supplementary materials.
    """
    return (beta / cluster_size) * (
        alphas["prot"] * n_prot + alphas["nuc"] * n_nuc + alphas["ligand"] * n_ligand
    )


@typecheck
def get_chain_weight(
    molecule_type: CLUSTERING_RESIDUE_MOLECULE_TYPE,
    cluster_size: int,
    alphas: Dict[str, float],
    beta: float,
) -> float:
    """Calculates the weight of a chain based on its type."""
    n_prot, n_nuc, n_ligand = get_chain_count(molecule_type)
    return calculate_weight(alphas, beta, n_prot, n_nuc, n_ligand, cluster_size)


@typecheck
def get_interface_weight(
    molecule_type_1: CLUSTERING_RESIDUE_MOLECULE_TYPE,
    molecule_type_2: CLUSTERING_RESIDUE_MOLECULE_TYPE,
    cluster_size: int,
    alphas: Dict[str, float],
    beta: float,
) -> float:
    """Calculates the weight of an interface based on the types of the two molecules."""
    p1, n1, l1 = get_chain_count(molecule_type_1)
    p2, n2, l2 = get_chain_count(molecule_type_2)

    n_prot = p1 + p2
    n_nuc = n1 + n2
    n_ligand = l1 + l2

    return calculate_weight(alphas, beta, n_prot, n_nuc, n_ligand, cluster_size)


@typecheck
def get_cluster_sizes(
    mapping: pl.DataFrame,
    cluster_id_col: str,
) -> Dict[int, int]:
    """
    Returns a dictionary where keys are cluster IDs and values are the number
    of chains/interfaces in the cluster.
    """
    cluster_sizes = mapping.group_by(cluster_id_col).agg(pl.len()).sort(cluster_id_col)
    return {row[0]: row[1] for row in cluster_sizes.iter_rows()}


@typecheck
def compute_chain_weights(
    chains: pl.DataFrame, alphas: Dict[str, float], beta: float
) -> pl.Series:
    """Computes the weights of the chains based on the cluster sizes."""
    molecule_idx = chains.get_column_index("molecule_id")
    cluster_idx = chains.get_column_index("cluster_id")
    cluster_sizes = get_cluster_sizes(chains, "cluster_id")

    return (
        chains.map_rows(
            lambda row: get_chain_weight(
                row[molecule_idx].split("-")[0],
                cluster_sizes[row[cluster_idx]],
                alphas,
                beta,
            ),
            return_dtype=pl.Float32,
        )
        .to_series(0)
        .rename("weight")
    )


@typecheck
def compute_interface_weights(
    interfaces: pl.DataFrame, alphas: Dict[str, float], beta: float
) -> pl.Series:
    """Computes the weights of the interfaces based on the chain weights."""
    molecule_idx_1 = interfaces.get_column_index("interface_molecule_id_1")
    molecule_idx_2 = interfaces.get_column_index("interface_molecule_id_2")
    cluster_idx = interfaces.get_column_index("interface_cluster_id")
    cluster_sizes = get_cluster_sizes(interfaces, "interface_cluster_id")

    return (
        interfaces.map_rows(
            lambda row: get_interface_weight(
                row[molecule_idx_1].split("-")[0],
                row[molecule_idx_2].split("-")[0],
                cluster_sizes[row[cluster_idx]],
                alphas,
                beta,
            ),
            return_dtype=pl.Float32,
        )
        .to_series(0)
        .rename("weight")
    )


class WeightedPDBSampler(Sampler[List[str]]):
    """
    Initializes a sampler for weighted sampling of PDB and chain/interface IDs.

    :param chain_mapping_paths: Path to the CSV file containing chain cluster
        mappings. If multiple paths are provided, they will be concatenated.
    :param interface_mapping_path: Path to the CSV file containing interface
        cluster mappings.
    :param batch_size: Number of PDB IDs to sample in each batch.
    :param beta_chain: Weighting factor for chain clusters.
    :param beta_interface: Weighting factor for interface clusters.
    :param alpha_prot: Weighting factor for protein chains.
    :param alpha_nuc: Weighting factor for nucleic acid chains.
    :param alpha_ligand: Weighting factor for ligand chains.
    :param pdb_ids_to_skip: List of PDB IDs to skip during sampling.
        Allow extra data filtering to ensure we avoid training
        on anomolous complexes that passed through all filtering
        and clustering steps.
    :param subset_to_ids: An optional list of mapping DataFrame indices
        to which to subset the original combined mapping DataFrame. This
        is primarily useful for debugging using a smaller set of clusters.

    Example:
    ```
    sampler = WeightedPDBSampler(...)
    for batch in sampler:
        print(batch)
    ```
    """

    def __init__(
        self,
        chain_mapping_paths: str | List[str],
        interface_mapping_path: str,
        batch_size: int,
        beta_chain: float = 0.5,
        beta_interface: float = 1.0,
        alpha_prot: float = 3.0,
        alpha_nuc: float = 3.0,
        alpha_ligand: float = 1.0,
        pdb_ids_to_skip: List[str] = [],
        pdb_ids_to_keep: list[str] | None = None,
    ):
        # Load chain and interface mappings
        if not isinstance(chain_mapping_paths, list):
            chain_mapping_paths = [chain_mapping_paths]

        chain_mapping = [pl.read_csv(path) for path in chain_mapping_paths]
        # Increment chain cluster IDs to avoid overlap
        chain_cluster_nums = [mapping.get_column("cluster_id").max() for mapping in chain_mapping]
        for i in range(1, len(chain_mapping)):
            chain_mapping[i] = chain_mapping[i].with_columns(
                (pl.col("cluster_id") + sum(chain_cluster_nums[:i])).alias("cluster_id")
            )

        chain_mapping = pl.concat(chain_mapping)
        interface_mapping = pl.read_csv(interface_mapping_path)

        # Filter out unwanted PDB IDs
        if len(pdb_ids_to_skip) > 0:
            chain_mapping = chain_mapping.filter(pl.col("pdb_id").is_in(pdb_ids_to_skip).not_())
            interface_mapping = interface_mapping.filter(
                pl.col("pdb_id").is_in(pdb_ids_to_skip).not_()
            )

        # Calculate weights for chains and interfaces
        self.alphas = {"prot": alpha_prot, "nuc": alpha_nuc, "ligand": alpha_ligand}
        self.betas = {"chain": beta_chain, "interface": beta_interface}
        self.batch_size = batch_size

        logger.info(
            "Precomputing chain and interface weights. This may take several minutes to complete."
        )

        # Subset to specific PDB IDs if provided
        if exists(pdb_ids_to_keep):
            chain_mapping = chain_mapping.filter(pl.col("pdb_id").is_in(pdb_ids_to_keep)).select(
                ["pdb_id", "chain_id", "molecule_id", "cluster_id"]
            )
            interface_mapping = interface_mapping.filter(
                pl.col("pdb_id").is_in(pdb_ids_to_keep)
            ).select(
                [
                    "pdb_id",
                    "interface_chain_id_1",
                    "interface_chain_id_2",
                    "interface_molecule_id_1",
                    "interface_molecule_id_2",
                    "interface_chain_cluster_id_1",
                    "interface_chain_cluster_id_2",
                    "interface_cluster_id",
                ]
            )

        chain_mapping.insert_column(
            len(chain_mapping.columns),
            compute_chain_weights(chain_mapping, self.alphas, self.betas["chain"]),
        )
        interface_mapping.insert_column(
            len(interface_mapping.columns),
            compute_interface_weights(interface_mapping, self.alphas, self.betas["interface"]),
        )

        logger.info("Finished precomputing chain and interface weights.")

        # Concatenate chain and interface mappings
        chain_mapping = chain_mapping.with_columns(
            [
                pl.col("chain_id").alias("chain_id_1"),
                pl.lit("").alias("chain_id_2"),
            ]
        )
        chain_mapping = chain_mapping.select(
            ["pdb_id", "chain_id_1", "chain_id_2", "cluster_id", "weight"]
        )

        interface_mapping = interface_mapping.with_columns(
            [
                pl.col("interface_chain_id_1").alias("chain_id_1"),
                pl.col("interface_chain_id_2").alias("chain_id_2"),
                (
                    pl.col("interface_cluster_id") + chain_mapping.get_column("cluster_id").max()
                ).alias("cluster_id"),
            ]
        )
        interface_mapping = interface_mapping.select(
            ["pdb_id", "chain_id_1", "chain_id_2", "cluster_id", "weight"]
        )
        self.mappings = chain_mapping.extend(interface_mapping)

        # Normalize weights
        self.weights = self.mappings.get_column("weight").to_numpy()
        self.weights = self.weights / self.weights.sum()

    def __len__(self) -> int:
        """Returns the number of batches in the dataset."""
        return len(self.mappings) // self.batch_size

    def __iter__(self) -> Iterator[List[str]]:
        """Returns an iterator that yields batches of PDB IDs."""
        while True:
            sampled = self.sample(self.batch_size)

            for (
                pdb_id,
                _,
                _,
            ) in sampled:
                yield pdb_id

    @typecheck
    def sample(self, batch_size: int) -> List[Tuple[str, str, str]]:
        """Samples a chain ID or interface ID based on the weights of the chains/interfaces."""
        indices = np.random.choice(len(self.mappings), size=batch_size, p=self.weights)
        return self.mappings[indices].select(["pdb_id", "chain_id_1", "chain_id_2"]).rows()

    @typecheck
    def cluster_based_sample(self, batch_size: int) -> List[Tuple[str, str, str]]:
        """
        Samples a chain ID or interface ID based on cluster IDs. For each batch, a number of cluster IDs
        are selected randomly, and a chain ID or interface ID is sampled from each cluster based on the
        weights of the chains/interfaces in the cluster.

        Warning! Significantly slower than the regular `sample` method.
        """
        cluster_ids = self.mappings.get_column("cluster_id").unique().sample(batch_size)

        pdb_chain_ids = []
        for cluster_id in cluster_ids:
            cluster = self.mappings.filter(pl.col("cluster_id") == cluster_id)
            if len(cluster) == 1:
                pdb_chain_ids.append(
                    cluster[0].select(["pdb_id", "chain_id_1", "chain_id_2"]).row(0)
                )
                continue
            cluster_weights = cluster.get_column("weight").to_numpy()
            cluster_weights = cluster_weights / cluster_weights.sum()
            idx = np.random.choice(len(cluster), p=cluster_weights)
            pdb_chain_ids.append(
                cluster[idx].select(["pdb_id", "chain_id_1", "chain_id_2"]).row(0)
            )

        return pdb_chain_ids


if __name__ == "__main__":
    interface_mapping_path = os.path.join(
        "data", "pdb_data", "data_caches", "clusterings", "interface_cluster_mapping.csv"
    )
    chain_mapping_paths = [
        os.path.join(
            "data", "pdb_data", "data_caches", "clusterings", "ligand_chain_cluster_mapping.csv"
        ),
        os.path.join(
            "data",
            "pdb_data",
            "data_caches",
            "clusterings",
            "nucleic_acid_chain_cluster_mapping.csv",
        ),
        os.path.join(
            "data", "pdb_data", "data_caches", "clusterings", "peptide_chain_cluster_mapping.csv"
        ),
        os.path.join(
            "data", "pdb_data", "data_caches", "clusterings", "protein_chain_cluster_mapping.csv"
        ),
    ]

    sampler = WeightedPDBSampler(
        chain_mapping_paths=chain_mapping_paths,
        interface_mapping_path=interface_mapping_path,
        batch_size=64,
    )

    print(sampler.sample(64))
    print(sampler.cluster_based_sample(64))



import os
from datetime import datetime
from loguru import logger
from typing import Any, Dict, List, Literal, Mapping, Tuple

import numpy as np
import polars as pl
import torch
import torch.nn.functional as F

from alphafold3_pytorch.common.biomolecule import (
    Biomolecule,
    _from_mmcif_object,
    get_residue_constants,
)
from alphafold3_pytorch.data import mmcif_parsing
from alphafold3_pytorch.life import (
    DNA_NUCLEOTIDES,
    HUMAN_AMINO_ACIDS,
    LIGANDS,
    RNA_NUCLEOTIDES,
)
from alphafold3_pytorch.utils.data_utils import extract_mmcif_metadata_field
from alphafold3_pytorch.utils.model_utils import (
    ExpressCoordinatesInFrame,
    RigidFrom3Points,
    distance_to_bins,
    get_frames_from_atom_pos,
)
from alphafold3_pytorch.tensor_typing import typecheck
from alphafold3_pytorch.utils.utils import exists

# Constants

TEMPLATE_TYPE = Literal["protein", "dna", "rna"]


@typecheck
def parse_m8(
    m8_filepath: str,
    template_type: TEMPLATE_TYPE,
    query_id: str,
    mmcif_dir: str,
    max_templates: int | None = None,
    num_templates: int | None = None,
    template_cutoff_date: datetime | None = None,
    randomly_sample_num_templates: bool = False,
    verbose: bool = False,
) -> List[Tuple[Biomolecule, TEMPLATE_TYPE]]:
    """Parse an M8 file and return a list of template Biomolecule objects.

    :param m8_filepath: The path to the M8 file.
    :param template_type: The type of template to parse.
    :param query_id: The ID of the query sequence.
    :param mmcif_dir: The directory containing mmCIF files.
    :param max_templates: The (optional) maximum number of templates to return.
    :param num_templates: The (optional) number of templates to return.
    :param template_cutoff_date: The (optional) cutoff date for templates.
    :param randomly_sample_num_templates: Whether to randomly sample the number of templates to
        return.
    :param verbose: Whether to log verbose output.
    :return: A list of template Biomolecule objects and their template types.
    """
    # Define the column names.
    columns = [
        "ID",
        "Template ID",
        "Identity",
        "Alignment Length",
        "Mismatches",
        "Gap Openings",
        "Query Start",
        "Query End",
        "Template Start",
        "Template End",
        "E-Value",
        "Bit Score",
        "Match String",
    ]

    # Read the M8 file as a DataFrame.
    try:
        df = pl.read_csv(m8_filepath, separator="\t", has_header=False, new_columns=columns)
    except Exception as e:
        if verbose:
            logger.warning(f"Skipping loading M8 file {m8_filepath} due to: {e}")
        return []

    # Filter the DataFrame to only include rows where
    # (1) the template ID does not contain any part of the query ID;
    # (2) the template's identity is between 0.1 and 0.95, exclusively;
    # (3) the alignment length is greater than 0;
    # (4) the template's length is at least 10; and
    # (5) the number of templates is less than the (optional) maximum number of templates.
    df = df.filter(~pl.col("Template ID").str.contains(query_id))
    df = df.filter((pl.col("Identity") > 0.1) & (pl.col("Identity") < 0.95))
    df = df.filter(pl.col("Alignment Length") > 0)
    df = df.filter((pl.col("Template End") - pl.col("Template Start")) >= 9)
    if exists(max_templates):
        df = df.head(max_templates)

    # Select the number of templates to return.
    if len(df) and exists(num_templates) and randomly_sample_num_templates:
        df = df.sample(min(len(df), num_templates))
    elif exists(num_templates):
        df = df.head(num_templates)

    # Load each template chain as a Biomolecule object.
    template_biomols = []
    for i in range(len(df)):
        row = df[i]
        row_template_id = row["Template ID"].item()
        template_id, template_chain = row_template_id.split("_")
        template_fpath = os.path.join(mmcif_dir, template_id[1:3], f"{template_id}-assembly1.cif")
        if not os.path.exists(template_fpath):
            continue
        try:
            template_mmcif_object = mmcif_parsing.parse_mmcif_object(
                template_fpath, row_template_id
            )
            template_release_date = extract_mmcif_metadata_field(
                template_mmcif_object, "release_date"
            )
            template_biomol = _from_mmcif_object(
                template_mmcif_object, chain_ids=set(template_chain)
            )
            if not (
                exists(template_cutoff_date)
                and datetime.strptime(template_release_date, "%Y-%m-%d") <= template_cutoff_date
            ):
                continue
            elif not exists(template_cutoff_date):
                pass
            if len(template_biomol.atom_positions):
                template_biomols.append((template_biomol, template_type))
        except Exception as e:
            if verbose:
                logger.warning(f"Skipping loading template {template_id} due to: {e}")

    return template_biomols


def _extract_template_features(
    template_biomol: Biomolecule,
    mapping: Mapping[int, int],
    template_sequence: str,
    query_sequence: str,
    query_chemtype: List[str],
    num_restype_classes: int = 32,
    num_distogram_bins: int = 39,
    distance_bins: List[float] = torch.linspace(3.25, 50.75, 38).float(),
    verbose: bool = False,
) -> Dict[str, Any]:
    """Parse atom positions in the target structure and align with the query.

    Atoms for each residue in the template structure are indexed to coincide
    with their corresponding residue in the query sequence, according to the
    alignment mapping provided.

    Adapted from:
    https://github.com/aqlaboratory/openfold/blob/main/openfold/data/templates.py

    :param template_biomol: `Biomolecule` representing the template.
    :param mapping: Dictionary mapping indices in the query sequence to indices in
        the template sequence.
    :param template_sequence: String describing the residue sequence for the
        template.
    :param query_sequence: String describing the residue sequence for the query.
    :param query_chemtype: List of strings describing the chemical type of each
        residue in the query sequence.
    :param num_restype_classes: The total number of residue types.
    :param num_dist_bins: The total number of distance bins.
    :param distance_bins: List of floats representing the bins for the distance
        histogram (i.e., distogram).
    :param verbose: Whether to log verbose output.

    :return: A dictionary containing the extra features derived from the template
        structure.
    """
    assert len(mapping) == len(query_sequence) == len(query_chemtype), (
        f"Mapping length {len(mapping)} must match query sequence length {len(query_sequence)} "
        f"and query chemtype length {len(query_chemtype)}."
    )
    assert num_distogram_bins == len(distance_bins) + 1, (
        f"Number of distance bins {num_distogram_bins} must match the length of distance bins "
        f"{len(distance_bins)} plus one."
    )

    all_atom_positions = template_biomol.atom_positions
    all_atom_mask = template_biomol.atom_mask

    all_atom_positions = np.split(all_atom_positions, all_atom_positions.shape[0])
    all_atom_masks = np.split(all_atom_mask, all_atom_mask.shape[0])

    template_restype = []
    template_all_atom_mask = []
    template_all_atom_positions = []

    template_distogram_atom_indices = []
    template_token_center_atom_indices = []
    template_three_atom_indices_for_frame = []

    for _, chemtype in zip(query_sequence, query_chemtype):
        # Handle residues in `query_sequence` that are not in `template_sequence`.
        query_chem_residue_constants = get_residue_constants(res_chem_index=chemtype)

        template_restype.append(query_chem_residue_constants.MSA_CHAR_TO_ID["-"])
        template_all_atom_mask.append(
            np.zeros(query_chem_residue_constants.atom_type_num, dtype=bool)
        )
        template_all_atom_positions.append(
            np.zeros((query_chem_residue_constants.atom_type_num, 3), dtype=np.float32)
        )

        template_distogram_atom_indices.append(0)
        template_token_center_atom_indices.append(0)
        template_three_atom_indices_for_frame.append(None)

    for query_index, template_index in mapping.items():
        # NOTE: Here, we assume that the query sequence's chemical types are the same as the
        # template sequence's chemical types. This is a reasonable assumption since the template
        # sequences are chemical type-specific search results for the query sequences.
        chemtype = query_chemtype[query_index]
        query_chem_residue_constants = get_residue_constants(res_chem_index=chemtype)

        if chemtype == 0:
            seq_mapping = HUMAN_AMINO_ACIDS
        elif chemtype == 1:
            seq_mapping = RNA_NUCLEOTIDES
        elif chemtype == 2:
            seq_mapping = DNA_NUCLEOTIDES
        elif chemtype == 3:
            seq_mapping = LIGANDS
        else:
            raise ValueError(f"Unrecognized chain chemical type: {chemtype}")

        if not (0 <= template_index < len(template_sequence)):
            if verbose:
                logger.warning(
                    f"Query index {query_index} is not mappable to the template sequence. "
                    f"Substituting with zero templates features for this position."
                )
            continue

        template_residue = template_sequence[template_index]
        template_res = seq_mapping.get(template_residue)

        # Handle modified polymer residues.
        is_polymer_res = chemtype < 3
        is_ligand_res = not is_polymer_res
        is_modified_polymer_res = is_polymer_res and (
            template_residue == "X"
            or template_residue not in query_chem_residue_constants.restype_1to3
        )
        if is_modified_polymer_res:
            template_res = LIGANDS.get("X")

        # Extract residue metadata.
        distogram_atom_idx = template_res["distogram_atom_idx"]
        token_center_atom_idx = template_res["token_center_atom_idx"]
        three_atom_indices_for_frame = template_res["three_atom_indices_for_frame"]

        if is_ligand_res or is_modified_polymer_res:
            # NOTE: Ligand and modified polymer residue representative atoms are located at
            # arbitrary indices, so we must use the atom mask to dynamically retrieve them.
            distogram_atom_idx = token_center_atom_idx = np.where(
                all_atom_masks[template_index][0]
            )[0][0]

        template_restype[query_index] = query_chem_residue_constants.MSA_CHAR_TO_ID.get(
            template_residue, query_chem_residue_constants.restype_num
        )
        template_all_atom_mask[query_index] = all_atom_masks[template_index][0]
        template_all_atom_positions[query_index] = all_atom_positions[template_index][0]

        template_distogram_atom_indices[query_index] = distogram_atom_idx
        template_token_center_atom_indices[query_index] = token_center_atom_idx
        template_three_atom_indices_for_frame[query_index] = three_atom_indices_for_frame

    # Assemble the template features tensors.
    template_restype = F.one_hot(torch.tensor(template_restype), num_classes=num_restype_classes)
    template_all_atom_mask = torch.from_numpy(np.stack(template_all_atom_mask))
    template_all_atom_positions = torch.from_numpy(np.stack(template_all_atom_positions))

    template_token_center_atom_indices = torch.tensor(template_token_center_atom_indices)
    template_token_center_atom_positions = torch.gather(
        template_all_atom_positions,
        1,
        template_token_center_atom_indices[..., None, None].expand(-1, -1, 3),
    ).squeeze(1)

    # Handle ligand and modified polymer residue frames.
    ligand_frames_present = not all(template_three_atom_indices_for_frame)
    template_backbone_frame_atom_mask = torch.ones(len(template_restype), dtype=torch.bool)
    if ligand_frames_present:
        template_token_center_atom_mask = torch.gather(
            template_all_atom_mask, 1, template_token_center_atom_indices.unsqueeze(-1)
        ).squeeze(1)
        new_frame_token_indices = get_frames_from_atom_pos(
            atom_pos=template_token_center_atom_positions,
            mask=template_token_center_atom_mask.bool(),
            filter_colinear_pos=True,
        )
        for token_index, frame_token_indices in enumerate(template_three_atom_indices_for_frame):
            if not exists(frame_token_indices):
                # Track invalid ligand frames.
                if (new_frame_token_indices[token_index] == -1).any():
                    template_backbone_frame_atom_mask[token_index] = False
                    template_three_atom_indices_for_frame[token_index] = (0, 0, 0)
                    continue

                # Collect the (token center) atom positions of the ligand frame atoms.
                new_frame_atom_positions = []
                new_frame_atom_mask = []
                for new_frame_token_index in new_frame_token_indices[token_index]:
                    new_frame_token_center_atom_index = template_token_center_atom_indices[
                        new_frame_token_index
                    ]
                    new_frame_atom_positions.append(
                        template_all_atom_positions[
                            new_frame_token_index, new_frame_token_center_atom_index
                        ].clone()
                    )
                    new_frame_atom_mask.append(
                        template_all_atom_mask[
                            new_frame_token_index, new_frame_token_center_atom_index
                        ].clone()
                    )

                # Move the ligand frame atoms to the first three positions of the
                # ligand residue's atom positions tensor.
                for local_new_frame_atom_index in range(len(new_frame_token_indices[token_index])):
                    template_all_atom_positions[
                        token_index, local_new_frame_atom_index
                    ] = new_frame_atom_positions[local_new_frame_atom_index]
                    template_all_atom_mask[
                        token_index, local_new_frame_atom_index
                    ] = new_frame_atom_mask[local_new_frame_atom_index]

                # Update ligand metadata after moving the frame atoms.
                template_distogram_atom_indices[token_index] = 1
                template_token_center_atom_indices[token_index] = 1
                template_three_atom_indices_for_frame[token_index] = (0, 1, 2)

    # Assemble the distogram and frame atom index tensors.
    template_distogram_atom_indices = torch.tensor(template_distogram_atom_indices)
    template_three_atom_indices_for_frame = torch.tensor(template_three_atom_indices_for_frame)

    # Construct pseudo beta mask.
    template_pseudo_beta_mask = torch.gather(
        template_all_atom_mask, 1, template_distogram_atom_indices.unsqueeze(-1)
    ).squeeze(-1)

    # Construct backbone frame mask.
    template_backbone_frame_mask = template_backbone_frame_atom_mask & torch.gather(
        template_all_atom_mask,
        1,
        template_three_atom_indices_for_frame,
    ).all(-1)

    # Construct distogram.
    template_distogram_atom_positions = torch.gather(
        template_all_atom_positions,
        1,
        template_distogram_atom_indices[..., None, None].expand(-1, -1, 3),
    ).squeeze(1)
    template_distogram_dist = torch.cdist(
        template_distogram_atom_positions, template_distogram_atom_positions, p=2
    )
    template_distogram_dist_binned = distance_to_bins(template_distogram_dist, distance_bins)

    template_distogram_dist_binned[
        # NOTE: This assigns the last bin to distances greater than the maximum bin (e.g., > 50.75 Å).
        template_distogram_dist
        > distance_bins.max()
    ] = (num_distogram_bins - 1)

    template_distogram = F.one_hot(
        template_distogram_dist_binned, num_classes=num_distogram_bins
    ).float()

    # Construct unit vectors.
    template_unit_vector = torch.zeros(
        (len(template_restype), len(template_restype), 3), dtype=torch.float32
    )

    template_backbone_frame_atom_positions = torch.gather(
        template_all_atom_positions,
        1,
        template_three_atom_indices_for_frame.unsqueeze(-1).expand(-1, -1, 3),
    )

    rigid_from_three_points = RigidFrom3Points()
    template_backbone_frames, _ = rigid_from_three_points(
        template_backbone_frame_atom_positions.unbind(-2)
    )

    express_coordinates_in_frame = ExpressCoordinatesInFrame()
    template_unit_vector = express_coordinates_in_frame(
        template_token_center_atom_positions.unsqueeze(0),
        template_backbone_frames.unsqueeze(0),
        pairwise=True,
    ).squeeze(0)

    return {
        "template_restype": template_restype.float(),
        "template_pseudo_beta_mask": template_pseudo_beta_mask.bool(),
        "template_backbone_frame_mask": template_backbone_frame_mask,
        "template_distogram": template_distogram,
        "template_unit_vector": template_unit_vector,
    }


class QueryToTemplateAlignError(Exception):
    """An error indicating that the query can't be aligned to the template."""



"""An mmCIF file format parser."""

import dataclasses
import functools
import io
import itertools
import logging
from collections import defaultdict
from operator import itemgetter
from typing import Any, Mapping, Optional, Sequence, Set, Tuple

import numpy as np
from Bio import PDB
from Bio.Data import PDBData

from alphafold3_pytorch.utils.data_utils import is_polymer, is_water, matrix_rotate

# Type aliases:
ChainId = str
PdbHeader = Mapping[str, Any]
PdbStructure = PDB.Structure.Structure
SeqRes = str
MmCIFDict = Mapping[str, Sequence[str]]

AtomFullId = Tuple[str, int, str, Tuple[str, int, str], Tuple[str, str]]
ResidueFullId = Tuple[str, int, str, Tuple[str, int, str]]
ChainFullId = Tuple[str, int, str]


@dataclasses.dataclass(frozen=True)
class Monomer:
    """Represents a monomer in a polymer chain."""

    id: str
    num: int


@dataclasses.dataclass(frozen=True)
class ChemComp:
    """Represents a chemical composition."""

    id: str
    formula: str
    formula_weight: str
    mon_nstd_flag: str
    name: str
    type: str


# Note - mmCIF format provides no guarantees on the type of author-assigned
# sequence numbers. They need not be integers.
@dataclasses.dataclass(frozen=True)
class AtomSite:
    """Represents an atom site in an mmCIF file."""

    mmcif_entity_num: int
    residue_name: str
    author_chain_id: str
    mmcif_chain_id: str
    author_seq_num: str
    mmcif_seq_num: int
    insertion_code: str
    hetatm_atom: str
    model_num: int


@dataclasses.dataclass(frozen=True)
class Bond:
    """Represents a structural connection between two atoms."""

    ptnr1_auth_seq_id: str
    ptnr1_auth_comp_id: str
    ptnr1_auth_asym_id: str
    ptnr1_label_atom_id: str
    pdbx_ptnr1_label_alt_id: str

    ptnr2_auth_seq_id: str
    ptnr2_auth_comp_id: str
    ptnr2_auth_asym_id: str
    ptnr2_label_atom_id: str
    pdbx_ptnr2_label_alt_id: str

    pdbx_leaving_atom_flag: str
    pdbx_dist_value: str
    pdbx_role: str
    conn_type_id: str


# Used to map SEQRES index to a residue in the structure.
@dataclasses.dataclass(frozen=True)
class ResiduePosition:
    """Represents a residue position in a chain."""

    chain_id: str
    residue_number: int
    insertion_code: str


@dataclasses.dataclass(frozen=True)
class ResidueAtPosition:
    """Represents a residue at a given position in a chain."""

    position: Optional[ResiduePosition]
    name: str
    is_missing: bool
    hetflag: str


@dataclasses.dataclass(frozen=True)
class MmcifObject:
    """Representation of a parsed mmCIF file.

    Contains:
        file_id: A meaningful name, e.g. a pdb_id. Should be unique amongst all
            files being processed.
        header: Biopython header.
        structure: Biopython structure.
        chem_comp_details: Dict mapping chain_id to a list of (non-missing) ChemComp. E.g.
            {'A': [ChemComp, ChemComp, ...]}
        all_chem_comp_details: Dict mapping chain_id to a list of ChemComp. E.g.
            {'A': [ChemComp, ChemComp, ...]}
        chain_to_seqres: Dict mapping chain_id to 1 letter sequence. E.g.
            {'A': 'ABCDEFG'}
        seqres_to_structure: Dict; for each chain_id contains a mapping between
            SEQRES index and a ResidueAtPosition. e.g. {'A': {0: ResidueAtPosition,
                                                            1: ResidueAtPosition,
                                                            ...}}
        entity_to_chain: Dict mapping entity_id to a list of chain_ids. E.g.
            {1: ['A', 'B']}
        mmcif_to_author_chain: Dict mapping internal mmCIF chain ids to author chain ids. E.g.
            {'A': 'B', 'B', 'B'}
        bonds: List of Bond objects.
        raw_string: The raw string used to construct the MmcifObject.
        atoms_to_remove: Optional set of atoms to remove.
        residues_to_remove: Optional set of residues to remove.
        chains_to_remove: Optional set of chains to remove.
    """

    file_id: str
    header: PdbHeader
    structure: PdbStructure
    chem_comp_details: Mapping[ChainId, Sequence[ChemComp]]
    all_chem_comp_details: Mapping[ChainId, Sequence[ChemComp]]
    chain_to_seqres: Mapping[ChainId, SeqRes]
    seqres_to_structure: Mapping[ChainId, Mapping[int, ResidueAtPosition]]
    entity_to_chain: Mapping[int, Sequence[str]]
    mmcif_to_author_chain: Mapping[str, str]
    bonds: Sequence[Bond]
    raw_string: Any
    atoms_to_remove: Set[AtomFullId]
    residues_to_remove: Set[ResidueFullId]
    chains_to_remove: Set[ChainFullId]


@dataclasses.dataclass(frozen=True)
class ParsingResult:
    """Returned by the parse function.

    Contains:
      mmcif_object: A MmcifObject, may be None if no chain could be successfully
        parsed.
      errors: A dict mapping (file_id, chain_id) to any exception generated.
    """

    mmcif_object: Optional[MmcifObject]
    errors: Mapping[Tuple[str, str], Any]


class ParseError(Exception):
    """An error indicating that an mmCIF file could not be parsed."""


def mmcif_loop_to_list(prefix: str, parsed_info: MmCIFDict) -> Sequence[Mapping[str, str]]:
    """Extracts loop associated with a prefix from mmCIF data as a list.

    Reference for loop_ in mmCIF:
      http://mmcif.wwpdb.org/docs/tutorials/mechanics/pdbx-mmcif-syntax.html

    :param prefix: Prefix shared by each of the data items in the loop.
        e.g. '_entity_poly_seq.', where the data items are _entity_poly_seq.num,
        _entity_poly_seq.mon_id. Should include the trailing period.
    :param parsed_info: A dict of parsed mmCIF data, e.g. _mmcif_dict from a Biopython
        parser.

    :return: A list of dicts; each dict represents 1 entry from an mmCIF loop.
    """
    cols = []
    data = []
    for key, value in parsed_info.items():
        if key.startswith(prefix):
            cols.append(key)
            data.append(value)

    assert all([len(xs) == len(data[0]) for xs in data]), (
        "mmCIF error: Not all loops are the same length: %s" % cols
    )

    return [dict(zip(cols, xs)) for xs in zip(*data)]


def mmcif_loop_to_dict(
    prefix: str,
    index: str,
    parsed_info: MmCIFDict,
) -> Mapping[str, Mapping[str, str]]:
    """Extracts loop associated with a prefix from mmCIF data as a dictionary.

    :param prefix: Prefix shared by each of the data items in the loop.
        e.g. '_entity_poly_seq.', where the data items are _entity_poly_seq.num,
        _entity_poly_seq.mon_id. Should include the trailing period.
    :param index: Which item of loop data should serve as the key.
    :param parsed_info: A dict of parsed mmCIF data, e.g. _mmcif_dict from a Biopython
        parser.

    :return: A dict of dicts; each dict represents 1 entry from an mmCIF loop,
      indexed by the index column.
    """
    entries = mmcif_loop_to_list(prefix, parsed_info)
    return {entry[index]: entry for entry in entries}


def apply_transformations(
    biomol: Any,
    transformation_dict: Mapping[str, Tuple[np.ndarray, np.ndarray]],
    operations: Sequence[Tuple[str, str]],
) -> Any:
    """
    Gets subassembly by applying the given operations to the input
    Biomolecule containing affected asym IDs.

    Adapted from: https://github.com/biotite-dev/biotite

    :param biomol: The Biomolecule to which to apply assembly transformations.
    :param transformation_dict: A dictionary, where the keys are operation IDs
        and the values are tuples of rotation matrix and translation vector.
    :param operations: A list of tuples, where each tuple contains operation IDs
        of a single step.
    :return: The subassembly Biomolecule.
    """
    # Additional first dimension for 'repeat_biomol()'
    assembly_coord = np.zeros((len(operations),) + biomol.atom_positions.shape)

    # Apply corresponding transformation for each copy in the assembly
    for i, operation in enumerate(operations):
        coord = biomol.atom_positions.copy()
        # Execute for each transformation step
        # in the operation expression
        for op_step in operation:
            rotation_matrix, translation_vector = transformation_dict[op_step]
            # Rotate
            coord = matrix_rotate(coord, rotation_matrix)
            # Translate
            coord += translation_vector
        assembly_coord[i] = coord

    return biomol.repeat(assembly_coord)


def get_transformations(
    struct_oper: Mapping[str, str]
) -> Mapping[str, Tuple[np.ndarray, np.ndarray]]:
    """
    Gets transformation operations in terms of rotation matrix and
    translation for each operation ID in `pdbx_struct_oper_list`.

    Adapted from: https://github.com/biotite-dev/biotite

    :param struct_oper: A dictionary representing each row in the
        `_pdbx_struct_oper_list` category of an mmCIF file.
    :return: A dictionary, where the keys are operation IDs and the values
        are tuples of rotation matrix and translation vector.
    """
    transformation_dict = {}
    for id in struct_oper:
        rotation_matrix = np.array(
            [
                [struct_oper[id][f"_pdbx_struct_oper_list.matrix[{i}][{j}]"] for j in (1, 2, 3)]
                for i in (1, 2, 3)
            ],
            dtype=np.float64,
        )
        translation_vector = np.array(
            [struct_oper[id][f"_pdbx_struct_oper_list.vector[{i}]"] for i in (1, 2, 3)],
            dtype=np.float64,
        )
        transformation_dict[id] = (rotation_matrix, translation_vector)
    return transformation_dict


def parse_operation_expression(expression: str) -> Sequence[Tuple[str, str]]:
    """
    Gets successive operation steps (IDs) for the given
    ``oper_expression``. Forms the cartesian product, if necessary.

    Adapted from: https://github.com/biotite-dev/biotite

    :param expression: The operation expression string.
    :return: A list of tuples, where each tuple contains operation IDs
        of a single step.
    """
    # Split groups by parentheses:
    # use the opening parenthesis as delimiter
    # and just remove the closing parenthesis
    # example: '(X0)(1-10,21-25)' from PDB `1a34`
    expressions_per_step = expression.replace(")", "").split("(")
    expressions_per_step = [e for e in expressions_per_step if len(e) > 0]
    # Important: Operations are applied from right to left
    expressions_per_step.reverse()

    operations = []
    for one_step_expr in expressions_per_step:
        one_step_op_ids = []
        for expr in one_step_expr.split(","):
            if "-" in expr:
                # Range of operation IDs, they must be integers
                first, last = expr.split("-")
                one_step_op_ids.extend([str(id) for id in range(int(first), int(last) + 1)])
            else:
                # Single operation ID
                one_step_op_ids.append(expr)
        operations.append(one_step_op_ids)

    # Cartesian product of operations
    return list(itertools.product(*operations))


@functools.lru_cache(16, typed=False)
def parse(
    *,
    file_id: str,
    mmcif_string: str,
    catch_all_errors: bool = True,
    auth_chains: bool = True,
    auth_residues: bool = True,
) -> ParsingResult:
    """Entry point, parses an mmcif_string.

    :param file_id: A string identifier for this file. Should be unique within the
        collection of files being processed.
    :param mmcif_string: Contents of an mmCIF file.
    :param catch_all_errors: If True, all exceptions are caught and error messages are
        returned as part of the ParsingResult. If False exceptions will be allowed
        to propagate.
    :param auth_chains: If True, use author-assigned chain ids. If False, use internal
        mmCIF chain ids.
    :param auth_residues: If True, use author-assigned residue numbers. If False, use
        internal mmCIF residue numbers.

    :return: A ParsingResult.
    """
    errors = {}
    try:
        parser = PDB.MMCIFParser(QUIET=True, auth_chains=auth_chains, auth_residues=auth_residues)
        with io.StringIO(mmcif_string) as handle:
            full_structure = parser.get_structure("", handle)
        first_model_structure = _get_first_model(full_structure)
        # Extract the _mmcif_dict from the parser, which contains useful fields not
        # reflected in the Biopython structure.
        parsed_info = parser._mmcif_dict  # pylint:disable=protected-access

        # Ensure all values are lists, even if singletons.
        for key, value in parsed_info.items():
            if not isinstance(value, list):
                parsed_info[key] = [value]

        header = _get_header(parsed_info)

        # Determine the complex chains, and their start numbers according to the
        # internal mmCIF numbering scheme (likely but not guaranteed to be 1).
        valid_chains = _get_complex_chains(parsed_info=parsed_info)
        if not valid_chains:
            return ParsingResult(None, {(file_id, ""): "No complex chains found in this file."})
        mmcif_seq_start_num = {
            chain_id: min([monomer.num for monomer in seq])
            for chain_id, (seq, _) in valid_chains.items()
        }

        # Loop over the atoms for which we have coordinates. Populate a mapping:
        # -mmcif_to_author_chain_id (maps internal mmCIF chain ids to chain ids used
        # by the authors / Biopython).
        mmcif_to_author_chain_id = {}
        atom_site_list = _get_atom_site_list(parsed_info)
        for atom in atom_site_list:
            # NOTE: This is a potential bottleneck, which may be addressed by
            # instead referencing pairs of internal mmCIF chain IDs and author
            # chain IDs on a residue or chain-wise level. However, it does not
            # seem that such information is always available in a given mmCIF file.
            if atom.model_num != "1":
                # We only process the first model at the moment.
                continue
            mmcif_to_author_chain_id[atom.mmcif_chain_id] = atom.author_chain_id

        # Determine each (author) complex chain's start number
        # according to the author-assigned numbering scheme.
        author_seq_start_num = defaultdict(lambda: float("inf"))
        for chain_id, (seq, _) in valid_chains.items():
            author_chain = mmcif_to_author_chain_id[chain_id]
            author_seq_start_num[author_chain] = min(
                author_seq_start_num[author_chain],
                min([monomer.num for monomer in seq]),
            )

        # Loop over the atoms for which we have coordinates. Populate two mappings:
        # -mmcif_seq_to_structure_mappings (maps mmCIF idx into sequence to author ResidueAtPosition).
        # -mmcif_entity_to_author_chain_mappings (maps entity number to author chain IDs).
        mmcif_seq_to_structure_mappings = {}
        mmcif_entity_to_author_chain_mappings = defaultdict(dict)
        for atom in atom_site_list:
            if atom.model_num != "1":
                # We only process the first model at the moment.
                continue

            if atom.mmcif_chain_id in valid_chains:
                hetflag = " "
                if atom.hetatm_atom == "HETATM":
                    # Water atoms are assigned a special hetflag of W in Biopython. We
                    # need to do the same, so that this hetflag can be used to fetch
                    # a residue from the Biopython structure by id.
                    if is_water(atom.residue_name):
                        hetflag = "W"
                    else:
                        hetflag = "H_" + atom.residue_name
                insertion_code = atom.insertion_code
                if not _is_set(atom.insertion_code):
                    insertion_code = " "
                position = ResiduePosition(
                    chain_id=atom.author_chain_id,
                    residue_number=int(atom.author_seq_num),
                    insertion_code=insertion_code,
                )
                mmcif_current = mmcif_seq_to_structure_mappings.get(atom.mmcif_chain_id, {})
                if _is_set(atom.mmcif_seq_num):
                    seq_idx = int(atom.mmcif_seq_num) - mmcif_seq_start_num[atom.mmcif_chain_id]
                else:
                    seq_idx = int(atom.author_seq_num) - author_seq_start_num[atom.author_chain_id]
                mmcif_current[seq_idx] = ResidueAtPosition(
                    position=position,
                    name=atom.residue_name,
                    is_missing=False,
                    hetflag=hetflag,
                )
                mmcif_seq_to_structure_mappings[atom.mmcif_chain_id] = mmcif_current
                mmcif_entity_to_author_chain_mappings[atom.mmcif_entity_num][
                    atom.author_chain_id
                ] = True

        # Add missing residue information to mmcif_seq_to_structure_mappings.
        for chain_id, (seq_info, chem_comp_info) in valid_chains.items():
            author_chain = mmcif_to_author_chain_id[chain_id]
            mmcif_current_mapping = mmcif_seq_to_structure_mappings[chain_id]
            for idx, monomer in enumerate(seq_info):
                seq_idx = idx + (mmcif_seq_start_num[chain_id] - 1)
                # NOTE: Water residues are often not labeled consecutively by authors
                # (e.g., see PDB 100d), so we avoid marking them as missing in this scenario.
                if seq_idx not in mmcif_current_mapping and not is_water(chem_comp_info[idx].id):
                    if not is_polymer(chem_comp_info[idx].type):
                        hetflag = "H_" + chem_comp_info[idx].id
                    else:
                        hetflag = " "
                    position = ResiduePosition(
                        chain_id=chain_id,
                        residue_number=seq_idx + 1,
                        insertion_code=" ",
                    )
                    mmcif_current_mapping[seq_idx] = ResidueAtPosition(
                        position=position,
                        name=monomer.id,
                        is_missing=True,
                        hetflag=hetflag,
                    )
            mmcif_seq_to_structure_mappings[chain_id] = dict(
                sorted(mmcif_current_mapping.items(), key=lambda x: x[0])
            )

        # Extract all sequence and chemical component details, and
        # populate seq_to_structure_mappings using author chain IDs.
        author_chain_to_sequence = defaultdict(str)
        all_chem_comp_details = defaultdict(list)
        seq_to_structure_mappings = defaultdict(dict)
        for chain_id, (seq_info, chem_comp_info) in valid_chains.items():
            author_chain = mmcif_to_author_chain_id[chain_id]
            current_mapping = seq_to_structure_mappings[author_chain]
            mmcif_current_mapping = mmcif_seq_to_structure_mappings[chain_id]
            seq = []
            all_chem_comp_info = []
            for monomer_index, monomer in enumerate(seq_info):
                all_chem_comp_info.append(chem_comp_info[monomer_index])
                if "peptide" in chem_comp_info[monomer_index].type.lower():
                    code = PDBData.protein_letters_3to1.get(f"{monomer.id: <3}", "X")
                elif (
                    "dna" in chem_comp_info[monomer_index].type.lower()
                    or "rna" in chem_comp_info[monomer_index].type.lower()
                ):
                    code = PDBData.nucleic_letters_3to1.get(f"{monomer.id: <3}", "X")
                else:
                    code = "X"
                seq.append(code if len(code) == 1 else "X")
            author_chain_to_sequence[author_chain] += "".join(seq)
            all_chem_comp_details[author_chain].extend(all_chem_comp_info)
            if current_mapping:
                start_index = len(current_mapping)
                current_mapping.update(
                    {
                        start_index + value_index: value
                        for value_index, value in enumerate(mmcif_current_mapping.values())
                    }
                )
            else:
                current_mapping.update(mmcif_current_mapping)

        # NOTE: All three of the following variables need to be perfectly matching
        # in terms of sequence contents to guarantee correctness for downstream code.
        for author_chain in author_chain_to_sequence:
            # Ensure the `seq_to_structure_mappings` is zero-indexed and does not contain index gaps.
            seq_to_structure_mappings[author_chain] = {
                i: value
                for i, value in enumerate(seq_to_structure_mappings[author_chain].values())
            }
            assert (
                len(author_chain_to_sequence[author_chain])
                == len(all_chem_comp_details[author_chain])
                == len(seq_to_structure_mappings[author_chain])
            ), (
                f"In parse(), encountered a sequence length mismatch for chain {author_chain} of file {file_id} regarding `author_chain_to_sequence`, `all_chem_comp_details`, and `seq_to_structure_mapping`: "
                f"{len(author_chain_to_sequence[author_chain])} != "
                f"{len(all_chem_comp_details[author_chain])} != "
                f"{len(seq_to_structure_mappings[author_chain])}"
            )

        # Identify only chemical component details that are present in the structure.
        chem_comp_details = {
            chain_id: [
                chem_comp_info
                for res_index, chem_comp_info in enumerate(all_chem_comp_details[chain_id])
                if not seq_to_structure_mappings[chain_id][res_index].is_missing
            ]
            for chain_id in all_chem_comp_details
        }

        # Simplify entity-to-chain mapping.
        entity_to_chain = {
            entity_id: list(chains.keys())
            for entity_id, chains in mmcif_entity_to_author_chain_mappings.items()
        }

        # Identify all bonds.
        bonds = _get_bond_list(parsed_info)

        mmcif_object = MmcifObject(
            file_id=file_id,
            header=header,
            structure=first_model_structure,
            chem_comp_details=chem_comp_details,
            all_chem_comp_details=all_chem_comp_details,
            chain_to_seqres=author_chain_to_sequence,
            seqres_to_structure=seq_to_structure_mappings,
            entity_to_chain=entity_to_chain,
            mmcif_to_author_chain=mmcif_to_author_chain_id,
            bonds=bonds,
            raw_string=parsed_info,
            atoms_to_remove=set(),
            residues_to_remove=set(),
            chains_to_remove=set(),
        )

        return ParsingResult(mmcif_object=mmcif_object, errors=errors)
    except Exception as e:  # pylint:disable=broad-except
        errors[(file_id, "")] = e
        if not catch_all_errors:
            raise
        return ParsingResult(mmcif_object=None, errors=errors)


def _get_first_model(structure: PdbStructure) -> PdbStructure:
    """Returns the first model in a Biopython structure."""
    return next(structure.get_models())


def get_release_date(parsed_info: MmCIFDict) -> str:
    """Returns the oldest revision date."""
    revision_dates = parsed_info["_pdbx_audit_revision_history.revision_date"]
    return min(revision_dates)


def _get_header(parsed_info: MmCIFDict) -> PdbHeader:
    """Returns a basic header containing method, release date and resolution."""
    header = {}

    experiments = mmcif_loop_to_list("_exptl.", parsed_info)
    header["structure_method"] = ",".join(
        [experiment["_exptl.method"].lower() for experiment in experiments]
    )

    # Note: The release_date here corresponds to the oldest revision. We prefer to
    # use this for dataset filtering over the deposition_date.
    if "_pdbx_audit_revision_history.revision_date" in parsed_info:
        header["release_date"] = get_release_date(parsed_info)
    else:
        logging.warning("Could not determine release_date for entry: %s", parsed_info["_entry.id"])

    header["resolution"] = 0.00
    for res_key in (
        "_refine.ls_d_res_high",
        "_em_3d_reconstruction.resolution",
        "_reflns.d_resolution_high",
    ):
        if res_key in parsed_info:
            try:
                raw_resolution = parsed_info[res_key][0]
                header["resolution"] = float(raw_resolution)
                break
            except ValueError:
                logging.debug("Invalid resolution format: %s", parsed_info[res_key])

    return header


def _get_atom_site_list(parsed_info: MmCIFDict) -> Sequence[AtomSite]:
    """Returns list of atom sites; contains data not present in the structure."""
    return [
        AtomSite(*site)
        for site in zip(  # pylint:disable=g-complex-comprehension
            parsed_info["_atom_site.label_entity_id"],
            parsed_info["_atom_site.label_comp_id"],
            parsed_info["_atom_site.auth_asym_id"],
            parsed_info["_atom_site.label_asym_id"],
            parsed_info["_atom_site.auth_seq_id"],
            parsed_info["_atom_site.label_seq_id"],
            parsed_info["_atom_site.pdbx_PDB_ins_code"],
            parsed_info["_atom_site.group_PDB"],
            parsed_info["_atom_site.pdbx_PDB_model_num"],
        )
    ]


def _get_bond_list(parsed_info: MmCIFDict) -> Sequence[Bond]:
    """Returns list of bonds present in the structure."""
    return [
        # Collect unique (partner) atom metadata required for each bond
        # per https://mmcif.wwpdb.org/docs/sw-examples/python/html/connections3.html.
        Bond(*conn)
        for conn in zip(  # pylint:disable=g-complex-comprehension
            # Partner 1
            parsed_info.get("_struct_conn.ptnr1_auth_seq_id", []),
            parsed_info.get("_struct_conn.ptnr1_auth_comp_id", []),
            parsed_info.get("_struct_conn.ptnr1_auth_asym_id", []),
            parsed_info.get("_struct_conn.ptnr1_label_atom_id", []),
            parsed_info.get("_struct_conn.pdbx_ptnr1_label_alt_id", []),
            # Partner 2
            parsed_info.get("_struct_conn.ptnr2_auth_seq_id", []),
            parsed_info.get("_struct_conn.ptnr2_auth_comp_id", []),
            parsed_info.get("_struct_conn.ptnr2_auth_asym_id", []),
            parsed_info.get("_struct_conn.ptnr2_label_atom_id", []),
            parsed_info.get("_struct_conn.pdbx_ptnr2_label_alt_id", []),
            # Connection metadata
            parsed_info.get("_struct_conn.pdbx_leaving_atom_flag", []),
            parsed_info.get("_struct_conn.pdbx_dist_value", []),
            parsed_info.get("_struct_conn.pdbx_role", []),
            parsed_info.get("_struct_conn.conn_type_id", []),
        )
        if len(conn[-1]) > 0
    ]


def _get_complex_chains(
    *, parsed_info: Mapping[str, Any]
) -> Mapping[ChainId, Tuple[Sequence[Monomer], Sequence[ChemComp]]]:
    """Extracts polymer information for complex chains.

    :param parsed_info: _mmcif_dict produced by the Biopython parser.

    :return: A dict mapping mmcif chain id to a tuple of a list of Monomers and a list of ChemComps.
    """
    # Get (non-)polymer information for each entity in the structure.
    poly_scheme = mmcif_loop_to_list("_pdbx_poly_seq_scheme.", parsed_info)
    branch_scheme = mmcif_loop_to_list("_pdbx_branch_scheme.", parsed_info)
    nonpoly_scheme = mmcif_loop_to_list("_pdbx_nonpoly_scheme.", parsed_info)
    # Get chemical compositions. Will allow us to identify which of these polymers
    # are protein, DNA, RNA, or ligand molecules.
    chem_comps = mmcif_loop_to_dict("_chem_comp.", "_chem_comp.id", parsed_info)

    polymers = defaultdict(list)
    for scheme in poly_scheme:
        polymers[scheme["_pdbx_poly_seq_scheme.asym_id"]].append(
            Monomer(
                id=scheme["_pdbx_poly_seq_scheme.mon_id"],
                num=int(scheme["_pdbx_poly_seq_scheme.seq_id"]),
            )
        )

    non_polymers = defaultdict(list)
    for scheme in branch_scheme:
        non_polymers[scheme["_pdbx_branch_scheme.asym_id"]].append(
            Monomer(
                id=scheme["_pdbx_branch_scheme.pdb_mon_id"],
                num=int(scheme["_pdbx_branch_scheme.pdb_seq_num"]),
            )
        )
    for scheme in nonpoly_scheme:
        non_polymers[scheme["_pdbx_nonpoly_scheme.asym_id"]].append(
            Monomer(
                id=scheme["_pdbx_nonpoly_scheme.pdb_mon_id"],
                num=int(scheme["_pdbx_nonpoly_scheme.pdb_seq_num"]),
            )
        )

    # Identify and return all complex chains.
    valid_chains = defaultdict(lambda: ([], []))
    for chain_id, seq_info in polymers.items():
        valid_chains[chain_id][0].extend(seq_info)
        valid_chains[chain_id][1].extend(
            [
                ChemComp(
                    id=chem_comps[monomer.id]["_chem_comp.id"],
                    formula=chem_comps[monomer.id]["_chem_comp.formula"],
                    formula_weight=chem_comps[monomer.id]["_chem_comp.formula_weight"],
                    mon_nstd_flag=chem_comps[monomer.id]["_chem_comp.mon_nstd_flag"],
                    name=chem_comps[monomer.id]["_chem_comp.name"],
                    type=chem_comps[monomer.id]["_chem_comp.type"],
                )
                for monomer in seq_info
            ]
        )

    # Insert non-polymer chains in-place into the valid_chains dict.
    for chain_id, seq_info in non_polymers.items():
        valid_chain = {
            monomer.num: (monomer, chem_comp)
            for monomer, chem_comp in zip(valid_chains[chain_id][0], valid_chains[chain_id][1])
        }
        new_valid_chain = {
            monomer.num: (monomer, chem_comp)
            for monomer, chem_comp in zip(
                seq_info,
                [
                    ChemComp(
                        id=chem_comps[monomer.id]["_chem_comp.id"],
                        formula=chem_comps[monomer.id]["_chem_comp.formula"],
                        formula_weight=chem_comps[monomer.id]["_chem_comp.formula_weight"],
                        mon_nstd_flag=chem_comps[monomer.id]["_chem_comp.mon_nstd_flag"],
                        name=chem_comps[monomer.id]["_chem_comp.name"],
                        type=chem_comps[monomer.id]["_chem_comp.type"],
                    )
                    for monomer in seq_info
                ],
            )
        }
        merged_valid_chain = {**valid_chain, **new_valid_chain}
        valid_chains[chain_id] = (
            [chain[0] for chain in merged_valid_chain.values()],
            [chain[1] for chain in merged_valid_chain.values()],
        )

    return valid_chains


def _is_set(data: str) -> bool:
    """Returns False if data is a special mmCIF character indicating 'unset'."""
    return data not in (".", "?")


def parse_mmcif_object(
    filepath: str, file_id: str, auth_chains: bool = True, auth_residues: bool = True
) -> MmcifObject:
    """Parse an mmCIF file into an `MmcifObject` containing a BioPython `Structure` object as well as associated metadata."""
    with open(filepath, "r") as f:
        mmcif_string = f.read()

    parsing_result = parse(
        file_id=file_id,
        mmcif_string=mmcif_string,
        auth_chains=auth_chains,
        auth_residues=auth_residues,
    )

    # Crash if an error is encountered. Any parsing errors should have
    # been dealt with beforehand (e.g., at the alignment stage).
    if parsing_result.mmcif_object is None:
        raise list(parsing_result.errors.values())[0]

    return parsing_result.mmcif_object


def filter_mmcif(mmcif_object: MmcifObject) -> MmcifObject:
    """Filter an `MmcifObject` based on collected (atom/residue/chain) removal sets."""
    model = mmcif_object.structure

    # Filter out specified chains
    chains_to_remove = set()

    for chain in model:
        # Filter out specified residues
        residues_to_remove = set()
        assert len(chain) == len(mmcif_object.chem_comp_details[chain.id]), (
            f"Number of residues in chain {chain.id} does not match "
            f"number of chemical component details for this chain: {len(chain)} vs. "
            f"{len(mmcif_object.chem_comp_details[chain.id])}."
        )
        for res_index, residue in enumerate(chain):
            # Filter out specified atoms
            atoms_to_remove = set()
            for atom in residue:
                if atom.get_full_id() in mmcif_object.atoms_to_remove:
                    atoms_to_remove.add(atom)
            if len(atoms_to_remove) == len(residue):
                residues_to_remove.add((res_index, residue))
            for atom in atoms_to_remove:
                residue.detach_child(atom.id)
            if residue.get_full_id() in mmcif_object.residues_to_remove:
                residues_to_remove.add((res_index, residue))
        if len(residues_to_remove) == len(chain):
            chains_to_remove.add(chain)
        for res_index, residue in sorted(residues_to_remove, key=itemgetter(0), reverse=True):
            del mmcif_object.chem_comp_details[chain.id][res_index]
            chain.detach_child(residue.id)
        if chain.get_full_id() in mmcif_object.chains_to_remove:
            chains_to_remove.add(chain)

    for chain in chains_to_remove:
        model.detach_child(chain.id)
        mmcif_object.chem_comp_details.pop(chain.id)

    mmcif_object.atoms_to_remove.clear()
    mmcif_object.residues_to_remove.clear()
    mmcif_object.chains_to_remove.clear()

    return mmcif_object



"""MSA loading functions used in AlphaFold2."""

# From: https://github.com/google-deepmind/alphafold/blob/f251de6613cb478207c732bf9627b1e853c99c2f/alphafold/data/parsers.py#L157

import dataclasses
import random
import re
import string
from typing import Literal, Optional, Sequence, Tuple

from alphafold3_pytorch.tensor_typing import typecheck

DeletionMatrix = Sequence[Sequence[int]]

# Constants

MSA_TYPE = Literal["protein", "dna", "rna"]

# Utilities for extracting identifiers from MSA sequence descriptions.


# Sequences coming from UniProtKB database come in the
# `db|UniqueIdentifier|EntryName` format, e.g. `tr|A0A146SKV9|A0A146SKV9_FUNHE`
# or `sp|P0C2L1|A3X1_LOXLA` (for TREMBL/Swiss-Prot respectively).
_UNIPROT_PATTERN = re.compile(
    r"""
    ^
    # UniProtKB/TrEMBL or UniProtKB/Swiss-Prot
    (?:tr|sp)
    \|
    # A primary accession number of the UniProtKB entry.
    (?P<AccessionIdentifier>[A-Za-z0-9]{6,10})
    # Occasionally there is a _0 or _1 isoform suffix, which we ignore.
    (?:_\d)?
    \|
    # TREMBL repeats the accession ID here. Swiss-Prot has a mnemonic
    # protein ID code.
    (?:[A-Za-z0-9]+)
    _
    # A mnemonic species identification code.
    (?P<SpeciesIdentifier>([A-Za-z0-9]){1,5})
    # Small BFD uses a final value after an underscore, which we ignore.
    (?:_\d+)?
    $
    """,
    re.VERBOSE,
)


@dataclasses.dataclass(frozen=True)
class Identifiers:
    species_id: str = ""


@typecheck
def _parse_sequence_identifier(msa_sequence_identifier: str) -> Identifiers:
    """Gets species from an msa sequence identifier.

    The sequence identifier has the format specified by
    _UNIPROT_TREMBL_ENTRY_NAME_PATTERN or _UNIPROT_SWISSPROT_ENTRY_NAME_PATTERN.
    An example of a sequence identifier: `tr|A0A146SKV9|A0A146SKV9_FUNHE`

    Args:
      msa_sequence_identifier: a sequence identifier.

    Returns:
      An `Identifiers` instance with species_id. These
      can be empty in the case where no identifier was found.
    """
    matches = re.search(_UNIPROT_PATTERN, msa_sequence_identifier.strip())
    if matches:
        return Identifiers(species_id=matches.group("SpeciesIdentifier"))
    return Identifiers()


@typecheck
def _extract_sequence_identifier(description: str) -> Optional[str]:
    """Extracts sequence identifier from description.

    Returns None if no match.
    """
    split_description = description.split()
    if split_description:
        return split_description[0].partition("/")[0]
    else:
        return None


@typecheck
def get_identifiers(description: str) -> Identifiers:
    """Computes extra MSA features from the description."""
    sequence_identifier = _extract_sequence_identifier(description)
    if sequence_identifier is None:
        return Identifiers()
    else:
        return _parse_sequence_identifier(sequence_identifier)


@dataclasses.dataclass(frozen=True)
class Msa:
    """Class representing a parsed MSA file."""

    sequences: Sequence[str]
    deletion_matrix: DeletionMatrix
    descriptions: Sequence[str]
    msa_type: MSA_TYPE

    def __post_init__(self):
        """Checks that all fields have the same length."""
        if not (len(self.sequences) == len(self.deletion_matrix) == len(self.descriptions)):
            raise ValueError(
                "All fields for an MSA must have the same length. "
                f"Got {len(self.sequences)} sequences, "
                f"{len(self.deletion_matrix)} rows in the deletion matrix, and "
                f"{len(self.descriptions)} descriptions."
            )

    def __len__(self):
        """Returns the number of sequences in the MSA."""
        return len(self.sequences)

    def truncate(self, max_seqs: int):
        """Truncates the MSA to the first `max_seqs` sequences."""
        max_seqs = min(len(self.sequences), max_seqs)
        return Msa(
            sequences=self.sequences[:max_seqs],
            deletion_matrix=self.deletion_matrix[:max_seqs],
            descriptions=self.descriptions[:max_seqs],
            msa_type=self.msa_type,
        )

    def random_truncate(self, max_seqs: int):
        """Truncates the MSA to a random range of `max_seqs` sequences."""
        max_seqs = min(len(self.sequences), max_seqs)
        start = random.randint(0, len(self.sequences) - max_seqs)  # nosec
        return Msa(
            sequences=self.sequences[start : start + max_seqs],
            deletion_matrix=self.deletion_matrix[start : start + max_seqs],
            descriptions=self.descriptions[start : start + max_seqs],
            msa_type=self.msa_type,
        )


@typecheck
def parse_fasta(fasta_string: str) -> Tuple[Sequence[str], Sequence[str]]:
    """Parses FASTA string and returns list of strings with amino-acid sequences.

    Arguments:
      fasta_string: The string contents of a FASTA file.

    Returns:
      A tuple of two lists:
      * A list of sequences.
      * A list of sequence descriptions taken from the comment lines. In the
        same order as the sequences.
    """
    sequences = []
    descriptions = []
    index = -1
    for line in fasta_string.splitlines():
        line = line.strip()
        if line.startswith(">"):
            index += 1
            descriptions.append(line[1:])  # Remove the '>' at the beginning.
            sequences.append("")
            continue
        elif not line:
            continue  # Skip blank lines.
        sequences[index] += line

    return sequences, descriptions


@typecheck
def parse_a3m(a3m_string: str, msa_type: MSA_TYPE) -> Msa:
    """Parses sequences and deletion matrix from a3m format alignment.

    Args:
      a3m_string: The string contents of a a3m file. The first sequence in the
        file should be the query sequence.
      msa_type: The type of the sequences in the MSA. This can be 'protein',
        'dna', or 'rna'.

    Returns:
      A tuple of:
        * A list of sequences that have been aligned to the query. These
          might contain duplicates.
        * The deletion matrix for the alignment as a list of lists. The element
          at `deletion_matrix[i][j]` is the number of residues deleted from
          the aligned sequence i at residue position j.
        * A list of descriptions, one per sequence, from the a3m file.
        * The type of the sequences in the MSA.
    """

    sequences, descriptions = parse_fasta(a3m_string)
    deletion_matrix = []
    for msa_sequence in sequences:
        deletion_vec = []
        deletion_count = 0
        for j in msa_sequence:
            if j.islower():
                deletion_count += 1
            else:
                deletion_vec.append(deletion_count)
                deletion_count = 0
        deletion_matrix.append(deletion_vec)

    # Make the MSA matrix out of aligned (deletion-free) sequences.
    deletion_table = str.maketrans("", "", string.ascii_lowercase)
    aligned_sequences = [s.translate(deletion_table) for s in sequences]
    return Msa(
        sequences=aligned_sequences,
        deletion_matrix=deletion_matrix,
        descriptions=descriptions,
        msa_type=msa_type,
    )



# Copyright 2021 AlQuraishi Laboratory
# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A Python wrapper for Kalign."""
import os
import subprocess  # nosec
import tempfile
from loguru import logger
from typing import Mapping, Sequence, Tuple

from alphafold3_pytorch.data import msa_parsing
from alphafold3_pytorch.data.template_parsing import (
    TEMPLATE_TYPE,
    QueryToTemplateAlignError,
)
from alphafold3_pytorch.utils.utils import exists


def _to_a3m(sequences: Sequence[str]) -> str:
    """Converts sequences to an a3m file."""
    names = ["sequence %d" % i for i in range(1, len(sequences) + 1)]
    a3m = []
    for sequence, name in zip(sequences, names):
        a3m.append(">" + name + "\n")
        a3m.append(sequence + "\n")
    return "".join(a3m)


def _realign_pdb_template_to_query(
    query_sequence: str,
    template_sequence: str,
    old_mapping: Mapping[int, int],
    kalign_binary_path: str | None,
    template_type: TEMPLATE_TYPE,
    min_frac_matching: float = 0.1,
    verbose: bool = False,
) -> Tuple[str, Mapping[int, int]]:
    """Aligns template sequence to the query sequence.

    Adapted from:
    https://github.com/aqlaboratory/openfold/blob/main/openfold/data/templates.py

    :param query_sequence: The actual query sequence extracted from a query
        Biomolecule object.
    :param template_sequence: The actual template sequence extracted from a template
        Biomolecule object.
    :param old_mapping: A mapping from the query sequence to the template sequence.
        This mapping will be used to compute the new mapping from the actual query
        sequence to the actual template sequence by aligning the `query_sequence`
        and the `template_sequence` strings.
    :param kalign_binary_path: The path to a kalign executable.
    :param template_type: The type of the template (e.g. "protein", "rna", "dna").
    :param min_frac_matching: The minimum fraction of matching residues between the
        actual template sequence and the actual query sequence. If the fraction is
        less than this value, a QueryToTemplateAlignError will be raised.
    :param verbose: Whether to log verbose output.

    :return: A tuple (new_template_sequence, new_query_to_template_mapping) where:
        * new_template_sequence is the actual template sequence that was given.
        * new_query_to_template_mapping is the new mapping from the query to the
            actual template sequence that was given.

    :raise: QueryToTemplateAlignError:
        * If there was an error thrown by the alignment tool.
        * Or if the actual template sequence differs (by default) by more than 90%
            from the `query_sequence` string.
    """
    assert exists(kalign_binary_path) and os.path.exists(
        kalign_binary_path
    ), f"Kalign binary not found at {kalign_binary_path}"
    aligner = Kalign(binary_path=kalign_binary_path)

    try:
        parsed_a3m = msa_parsing.parse_a3m(
            aligner.align([query_sequence, template_sequence]),
            template_type,
        )
        old_aligned_template, new_aligned_template = parsed_a3m.sequences
    except Exception as e:
        raise QueryToTemplateAlignError(
            "Could not align old template %s to template %s. Error: %s"
            % (
                query_sequence,
                template_sequence,
                str(e),
            )
        )

    if verbose:
        logger.info(
            "Old aligned template: %s\nNew aligned template: %s",
            old_aligned_template,
            new_aligned_template,
        )

    old_to_new_template_mapping = {}
    old_template_index = -1
    new_template_index = -1
    num_same = 0
    for old_template_res, new_template_res in zip(old_aligned_template, new_aligned_template):
        if old_template_res != "-":
            old_template_index += 1
        if new_template_res != "-":
            new_template_index += 1
        if old_template_res != "-" and new_template_res != "-":
            old_to_new_template_mapping[old_template_index] = new_template_index
            if old_template_res == new_template_res:
                num_same += 1

    # Require at least (by default) 10 % sequence identity w.r.t. to the shorter of the sequences.
    frac_matching = float(num_same) / min(len(query_sequence), len(template_sequence))
    if frac_matching < min_frac_matching:
        raise QueryToTemplateAlignError(
            "Insufficient similarity of the sequence in the database: %s to the "
            "actual sequence in the mmCIF file: %s. We require at least "
            f"{min_frac_matching * 100} %% similarity w.r.t. to the shorter of the sequences. This is not a "
            "problem unless you think this is a template that should be included."
            % (
                query_sequence,
                template_sequence,
            )
        )

    new_query_to_template_mapping = {}
    for query_index, old_template_index in old_mapping.items():
        new_query_to_template_mapping[query_index] = old_to_new_template_mapping.get(
            old_template_index, -1
        )

    template_sequence = template_sequence.replace("-", "")

    return template_sequence, new_query_to_template_mapping


class Kalign:
    """Python wrapper of the Kalign binary."""

    def __init__(self, *, binary_path: str):
        """Initializes the Python Kalign wrapper.

        Args:
          binary_path: The path to the Kalign binary.

        Raises:
          RuntimeError: If Kalign binary not found within the path.
        """
        self.binary_path = binary_path

    def align(self, sequences: Sequence[str]) -> str:
        """Aligns the sequences and returns the alignment in A3M string.

        Args:
          sequences: A list of query sequence strings. The sequences have to be at
            least 6 residues long (Kalign requires this). Note that the order in
            which you give the sequences might alter the output slightly as
            different alignment tree might get constructed.

        Returns:
          A string with the alignment in a3m format.

        Raises:
          RuntimeError: If Kalign fails.
          ValueError: If any of the sequences is less than 6 residues long.
        """
        logger.info("Aligning %d sequences", len(sequences))

        for s in sequences:
            if len(s) < 6:
                raise ValueError(
                    "Kalign requires all sequences to be at least 6 "
                    "residues long. Got %s (%d residues)." % (s, len(s))
                )

        with tempfile.TemporaryDirectory() as query_tmp_dir:
            input_fasta_path = os.path.join(query_tmp_dir, "input.fasta")
            output_a3m_path = os.path.join(query_tmp_dir, "output.a3m")

            with open(input_fasta_path, "w") as f:
                f.write(_to_a3m(sequences))

            cmd = [
                self.binary_path,
                "-i",
                input_fasta_path,
                "-o",
                output_a3m_path,
                "-format",
                "fasta",
            ]

            logger.info('Launching subprocess "%s"', " ".join(cmd))
            process = subprocess.Popen(
                cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )  # nosec

            stdout, stderr = process.communicate()
            retcode = process.wait()
            logger.info(
                "Kalign stdout:\n%s\n\nstderr:\n%s\n",
                stdout.decode("utf-8"),
                stderr.decode("utf-8"),
            )

            if retcode:
                raise RuntimeError(
                    "Kalign failed\nstdout:\n%s\n\nstderr:\n%s\n"
                    % (stdout.decode("utf-8"), stderr.decode("utf-8"))
                )

            with open(output_a3m_path) as f:
                a3m = f.read()

            return a3m



"""General-purpose data pipeline."""

import os
from typing import Dict, List, MutableMapping, Optional, Tuple

import numpy as np
import torch
import torch.nn.functional as F
from loguru import logger
from torch import Tensor

from alphafold3_pytorch.common.biomolecule import (
    Biomolecule,
    _from_mmcif_object,
    get_residue_constants,
    to_mmcif,
)
from alphafold3_pytorch.data import mmcif_parsing, msa_parsing
from alphafold3_pytorch.data.kalign import _realign_pdb_template_to_query
from alphafold3_pytorch.data.template_parsing import (
    TEMPLATE_TYPE,
    _extract_template_features,
)
from alphafold3_pytorch.tensor_typing import typecheck
from alphafold3_pytorch.utils.utils import exists

# Constants

GAP_ID = get_residue_constants("peptide").MSA_CHAR_TO_ID["-"]

FeatureDict = MutableMapping[str, np.ndarray | Tensor]


@typecheck
def make_sequence_features(sequence: str, description: str) -> FeatureDict:
    """Construct a feature dict of sequence features."""
    features = {}
    features["domain_name"] = np.array([description.encode("utf-8")], dtype=object)
    features["sequence"] = np.array([sequence.encode("utf-8")], dtype=object)
    return features


@typecheck
def make_msa_mask(features: FeatureDict) -> FeatureDict:
    """
    Make MSA mask features.
    From: https://github.com/aqlaboratory/openfold/blob/6f63267114435f94ac0604b6d89e82ef45d94484/openfold/data/data_transforms.py#L379
    NOTE: Openfold Mask features are all ones, but will later be zero-padded.

    :param features: The features dictionary.
    :return: The features dictionary with new MSA mask features.
    """
    features["msa_mask"] = torch.ones(features["msa"].shape[1], dtype=torch.float32)
    features["msa_row_mask"] = torch.ones((features["msa"].shape[0]), dtype=torch.float32)
    return features


@typecheck
def make_msa_features(
    msas: Dict[str, msa_parsing.Msa | None],
    chain_id_to_residue: Dict[str, Dict[str, List[int]]],
    ligand_chemtype_index: int = 3,
    raise_missing_exception: bool = False,
) -> FeatureDict:
    """
    Construct a feature dictionary of MSA features.
    From: https://github.com/aqlaboratory/openfold/blob/6f63267114435f94ac0604b6d89e82ef45d94484/openfold/data/data_pipeline.py#L224

    :param msas: The mapping of chain IDs to lists of (optional) MSAs for each chain.
    :param chain_id_to_residue: The mapping of chain IDs to residue information.
    :param ligand_index: The index of the ligand in the chemical type list.
    :param raise_missing_exception: Whether to raise an exception if no MSAs are provided for any chain.
    :return: The MSA feature dictionary.
    """
    if not msas:
        raise ValueError("At least one chain's MSA must be provided.")

    # Infer MSA metadata.
    max_alignments = 1
    for msa in msas.values():
        if exists(msa) and exists(msa.sequences) and exists(msa.sequences[0]):
            max_alignments = max(max_alignments, len(msa.sequences) if msa else 1)

    # Collect MSAs.
    int_msa_list = []
    deletion_matrix_list = []
    species_ids_list = []

    for chain_id, msa in msas.items():
        int_msa = []
        deletion_matrix = []
        species_ids = []
        seen_sequences = set()

        chain_chemtype = chain_id_to_residue[chain_id]["chemtype"]
        chain_residue_index = chain_id_to_residue[chain_id]["residue_index"]

        num_res = len(chain_chemtype)
        assert num_res == len(chain_residue_index), (
            f"Residue features count mismatch for chain {chain_id}: "
            f"{num_res} != {len(chain_residue_index)}"
        )

        msa_residue_constants = (
            get_residue_constants(msa.msa_type.replace("protein", "peptide"))
            if exists(msa)
            else None
        )

        gap_ids = [[GAP_ID] * num_res]
        deletion_values = [[0] * num_res]
        species = ["".encode("utf-8")]

        if not msa and raise_missing_exception:
            raise ValueError(f"MSA for chain {chain_id} must contain at least one sequence.")
        elif not msa:
            # Pad the MSA to the maximum number of alignments
            # if the chain does not have any associated alignments.
            int_msa_list.append(torch.tensor(gap_ids * max_alignments, dtype=torch.long))
            deletion_matrix_list.append(
                torch.tensor(deletion_values * max_alignments, dtype=torch.float32)
            )
            species_ids_list.append(np.array(species * max_alignments, dtype=object))
            continue

        for sequence_index, sequence in enumerate(msa.sequences):
            if sequence in seen_sequences:
                continue
            seen_sequences.add(sequence)

            # Convert the MSA to integers while handling
            # ligands and modified polymer residues.
            msa_res_types = []
            msa_deletion_values = []

            polymer_residue_index = -1

            for idx, (chemtype, residue_index) in enumerate(
                zip(chain_chemtype, chain_residue_index)
            ):
                is_polymer = chemtype < ligand_chemtype_index
                is_ligand = not is_polymer

                chem_residue_constants = get_residue_constants(res_chem_index=chemtype)

                # NOTE: For modified polymer residues, we only increment the polymer residue index
                # when the current (atomized) modified polymer residue's atom sequence ends.
                increment_index = (
                    0 < idx < num_res and chain_residue_index[idx - 1] != residue_index
                )
                polymer_residue_index += 1 if is_polymer and (idx == 0 or increment_index) else 0

                if is_ligand:
                    # NOTE: For ligands, we use the unknown amino acid type.
                    msa_res_type = chem_residue_constants.restype_num
                    msa_deletion_value = 0
                else:
                    # NOTE: For polymer residues of a different chemical type than the chain's MSA,
                    # we use the unknown residue type of the corresponding chemical type
                    # (e.g., `DN` for DNA residues in a protein chain's MSA). This should
                    # provide the model with partial information about the residue's identity.
                    if chem_residue_constants != msa_residue_constants:
                        msa_res_type = chem_residue_constants.restype_num
                    else:
                        res = sequence[polymer_residue_index]
                        msa_res_type = msa_residue_constants.MSA_CHAR_TO_ID.get(
                            res, msa_residue_constants.restype_num
                        )

                    msa_deletion_value = msa.deletion_matrix[sequence_index][polymer_residue_index]

                msa_res_types.append(msa_res_type)
                msa_deletion_values.append(msa_deletion_value)

            assert polymer_residue_index + 1 == len(
                sequence
            ), f"Polymer residue index length mismatch for MSA chain {chain_id}: {polymer_residue_index + 1} != {len(sequence)}"

            int_msa.append(msa_res_types)
            deletion_matrix.append(msa_deletion_values)

            identifiers = msa_parsing.get_identifiers(msa.descriptions[sequence_index])
            species_ids.append(identifiers.species_id.encode("utf-8"))

        # Pad the MSA to the maximum number of alignments.
        num_padding_alignments = max_alignments - len(int_msa)

        padding_msa = gap_ids * num_padding_alignments
        padding_deletion_matrix = deletion_values * num_padding_alignments
        padding_species_ids = species * num_padding_alignments

        int_msa_list.append(torch.tensor(int_msa + padding_msa, dtype=torch.long))
        deletion_matrix_list.append(
            torch.tensor(deletion_matrix + padding_deletion_matrix, dtype=torch.float32)
        )
        species_ids_list.append(np.array(species_ids + padding_species_ids, dtype=object))

    features = {
        "msa": torch.cat(int_msa_list, dim=-1),
        "deletion_matrix": torch.cat(deletion_matrix_list, dim=-1),
        "msa_species_identifiers": np.stack(species_ids_list),
        "num_alignments": max_alignments,
    }
    return features


@typecheck
def make_template_features(
    templates: Dict[str, List[Tuple[Biomolecule, TEMPLATE_TYPE]]],
    chain_id_to_residue: Dict[str, Dict[str, List[int]]],
    num_templates: int | None = None,
    kalign_binary_path: str | None = None,
    unknown_restype: int = 20,
    num_restype_classes: int = 32,
    num_distogram_bins: int = 39,
    raise_missing_exception: bool = False,
    verbose: bool = False,
) -> FeatureDict:
    """Construct a feature dictionary of template features.

    :param templates: The mapping of chain IDs to lists of template Biomolecule objects and their
        template types.
    :param chain_id_to_residue: The mapping of chain IDs to residue information.
    :param num_templates: The (optional) number of templates to return per chain.
    :param kalign_binary_path: The path to a local Kalign2 executable.
    :param unknown_restype: The unknown residue type index.
    :param num_restype_classes: The number of classes in the residue type classification.
    :param num_distogram_bins: The number of bins in the distogram features.
    :param raise_missing_exception: Whether to raise an exception if no templates are provided.
    :param verbose: Whether to log verbose output.
    :return: The template feature dictionary.
    """

    if exists(num_templates) and num_templates < 1:
        raise ValueError("The requested number of templates must be greater than zero.")

    # Infer template metadata.
    max_templates = num_templates if exists(num_templates) else 1
    for template in templates.values():
        if exists(template) and len(template):
            max_templates = max(max_templates, len(template))

    # Collect templates.
    template_restype_list = []
    template_pseudo_beta_mask_list = []
    template_backbone_frame_mask_list = []
    template_distogram_list = []
    template_unit_vector_list = []
    chain_index_list = []

    template_mask = torch.zeros(max_templates, dtype=torch.bool)

    for chain_index, (chain_id, template) in enumerate(templates.items()):
        chain_chemtype = chain_id_to_residue[chain_id]["chemtype"]
        chain_restype = chain_id_to_residue[chain_id]["restype"]
        chain_residue_index = chain_id_to_residue[chain_id]["residue_index"]

        # Map the chain's residue types to its query sequence.
        query_sequence = []
        for chemtype, restype in zip(chain_chemtype, chain_restype):
            chem_residue_constants = get_residue_constants(res_chem_index=chemtype)
            query_sequence.append(
                "X"
                if restype - chem_residue_constants.min_restype_num
                >= len(chem_residue_constants.restypes)
                else chem_residue_constants.restypes[
                    restype - chem_residue_constants.min_restype_num
                ]
            )
        query_sequence = "".join(query_sequence)

        num_res = len(chain_chemtype)
        assert num_res == len(chain_residue_index), (
            f"Residue features count mismatch for chain {chain_id}: "
            f"{num_res} != {len(chain_residue_index)}"
        )

        template_restype_list.append(
            F.one_hot(
                torch.full((max_templates, num_res), unknown_restype, dtype=torch.long),
                num_classes=num_restype_classes,
            )
        )
        template_pseudo_beta_mask_list.append(
            torch.zeros((max_templates, num_res), dtype=torch.bool)
        )
        template_backbone_frame_mask_list.append(
            torch.zeros((max_templates, num_res), dtype=torch.bool)
        )
        template_distogram_list.append(
            torch.zeros(
                (max_templates, num_res, num_res, num_distogram_bins),
                dtype=torch.float32,
            )
        )
        template_unit_vector_list.append(
            torch.zeros((max_templates, num_res, num_res, 3), dtype=torch.float32)
        )
        chain_index_list.append(torch.full((max_templates, num_res), chain_index))

        if not templates[chain_id] or not exists(kalign_binary_path):
            if raise_missing_exception:
                raise ValueError(
                    f"Templates for chain {chain_id} must contain at least one template and must be aligned with Kalign2."
                )
            continue

        for template_index, (template_biomol, template_type) in enumerate(template):
            template_chain_ids = list(
                dict.fromkeys(template_biomol.chain_id.tolist())
            )  # NOTE: we must maintain the order of unique chain IDs
            assert (
                len(template_chain_ids) == 1
            ), f"Template Biomolecule for chain {chain_id} must contain exactly one chain."
            template_chain_id = template_chain_ids[0]

            template_chain_chemtype = template_biomol.chemtype[
                template_biomol.chain_id == template_chain_id
            ].tolist()
            template_chain_restype = template_biomol.restype[
                template_biomol.chain_id == template_chain_id
            ].tolist()

            # Map the template chain's residue types to its target sequence.
            template_sequence = []
            for template_chemtype, template_restype in zip(
                template_chain_chemtype, template_chain_restype
            ):
                template_chem_residue_constants = get_residue_constants(
                    res_chem_index=template_chemtype
                )
                template_sequence.append(
                    "X"
                    if template_restype - template_chem_residue_constants.min_restype_num
                    >= len(template_chem_residue_constants.restypes)
                    else template_chem_residue_constants.restypes[
                        template_restype - template_chem_residue_constants.min_restype_num
                    ]
                )
            template_sequence = "".join(template_sequence)

            try:
                # Use Kalign to align the query and target sequences.
                mapping = {
                    x: x for x, curr_char in enumerate(query_sequence) if curr_char.isalnum()
                }
                realigned_template_sequence, realigned_mapping = _realign_pdb_template_to_query(
                    query_sequence=query_sequence,
                    template_sequence=template_sequence,
                    old_mapping=mapping,
                    kalign_binary_path=kalign_binary_path,
                    template_type=template_type,
                )

                # Extract features from aligned template.
                template_features = _extract_template_features(
                    template_biomol=template_biomol,
                    mapping=realigned_mapping,
                    template_sequence=realigned_template_sequence,
                    query_sequence=query_sequence,
                    query_chemtype=chain_chemtype,
                    num_restype_classes=num_restype_classes,
                    num_distogram_bins=num_distogram_bins,
                )

                template_restype_list[chain_index][template_index] = template_features[
                    "template_restype"
                ]
                template_pseudo_beta_mask_list[chain_index][template_index] = template_features[
                    "template_pseudo_beta_mask"
                ]
                template_backbone_frame_mask_list[chain_index][template_index] = template_features[
                    "template_backbone_frame_mask"
                ]
                template_distogram_list[chain_index][template_index] = template_features[
                    "template_distogram"
                ]
                template_unit_vector_list[chain_index][template_index] = template_features[
                    "template_unit_vector"
                ]
                template_mask[template_index] = True

            except Exception as e:
                if verbose:
                    logger.warning(
                        f"Skipping extraction of template features for chain {chain_id} due to: {e}"
                    )

                template_restype_list[chain_index][template_index] = F.one_hot(
                    torch.full((num_res,), unknown_restype, dtype=torch.long),
                    num_classes=num_restype_classes,
                )
                template_pseudo_beta_mask_list[chain_index][template_index] = torch.zeros(
                    num_res, dtype=torch.bool
                )
                template_backbone_frame_mask_list[chain_index][template_index] = torch.zeros(
                    num_res, dtype=torch.bool
                )
                template_distogram_list[chain_index][template_index] = torch.zeros(
                    (num_res, num_res, num_distogram_bins), dtype=torch.float32
                )
                template_unit_vector_list[chain_index][template_index] = torch.zeros(
                    (num_res, num_res, 3), dtype=torch.float32
                )

    # NOTE: Following AF3 Supplement, Section 2.4, the pairwise distogram and
    # unit vector features do not contain inter-chain interaction information.

    # Initialize an empty list to store the block_diag results.
    block_diag_distograms = []
    block_diag_unit_vectors = []

    # Loop over the first dimension (templates dimension).
    for i in range(max_templates):
        # Initialize an empty list to store each 2D block-diagonal matrix for distograms and unit vectors.
        distogram_blocks = []
        unit_vector_blocks = []

        # Loop over the last dimension (channels).
        for j in range(num_distogram_bins):
            # Extract the 2D slice and apply `block_diag()`.
            distogram_blocks.append(
                torch.block_diag(*[c[i, :, :, j] for c in template_distogram_list])
            )
            if j < 3:
                unit_vector_blocks.append(
                    torch.block_diag(*[c[i, :, :, j] for c in template_unit_vector_list])
                )

        # Stack along the third dimension (residues) and append to the list.
        block_diag_distograms.append(torch.stack(distogram_blocks, dim=-1))
        block_diag_unit_vectors.append(torch.stack(unit_vector_blocks, dim=-1))

    # Finalize template features by assembling each of them into a pairwise representation.
    template_backbone_frame_masks = torch.cat(template_backbone_frame_mask_list, dim=1)
    template_pseudo_beta_masks = torch.cat(template_pseudo_beta_mask_list, dim=1)
    template_distograms = torch.stack(block_diag_distograms, dim=0)
    template_unit_vectors = torch.stack(block_diag_unit_vectors, dim=0)
    template_restypes = torch.cat(template_restype_list, dim=1).float()
    chain_indices = torch.cat(chain_index_list, dim=1)

    template_backbone_pairwise_frame_masks = template_backbone_frame_masks.unsqueeze(
        1
    ) * template_backbone_frame_masks.unsqueeze(2)
    template_pseudo_beta_pairwise_masks = template_pseudo_beta_masks.unsqueeze(
        1
    ) * template_pseudo_beta_masks.unsqueeze(2)

    templates = torch.cat(
        (
            template_distograms,
            template_backbone_pairwise_frame_masks.unsqueeze(-1),
            template_unit_vectors,
            template_pseudo_beta_pairwise_masks.unsqueeze(-1),
        ),
        dim=-1,
    )

    is_same_chain = chain_indices.unsqueeze(1) == chain_indices.unsqueeze(2)
    templates *= is_same_chain.unsqueeze(-1)

    templates = torch.cat(
        (
            templates,
            template_restypes.unsqueeze(-3).expand(-1, template_restypes.shape[-2], -1, -1),
            template_restypes.unsqueeze(-2).expand(-1, -1, template_restypes.shape[-2], -1),
        ),
        dim=-1,
    )

    features = {
        "templates": templates,
        "template_mask": template_mask,
    }
    return features


@typecheck
def get_assembly(biomol: Biomolecule, assembly_id: Optional[str] = None) -> Biomolecule:
    """Get a specified (Biomolecule) assembly of a given Biomolecule.

    Adapted from: https://github.com/biotite-dev/biotite

    :param biomol: The Biomolecule from which to extract the requested assembly.
    :param assembly_id: The index of the assembly to get.
    :return: The requested assembly.
    """
    # Get mmCIF metadata categories
    assembly_category = mmcif_parsing.mmcif_loop_to_dict(
        "_pdbx_struct_assembly.", "_pdbx_struct_assembly.id", biomol.mmcif_metadata
    )
    assembly_gen_category = mmcif_parsing.mmcif_loop_to_dict(
        "_pdbx_struct_assembly_gen.",
        "_pdbx_struct_assembly_gen.assembly_id",
        biomol.mmcif_metadata,
    )
    struct_oper_category = mmcif_parsing.mmcif_loop_to_dict(
        "_pdbx_struct_oper_list.", "_pdbx_struct_oper_list.id", biomol.mmcif_metadata
    )

    if not all((assembly_category, assembly_gen_category, struct_oper_category)):
        logger.warning(
            "Not all required assembly information was found in the mmCIF file. Returning the input biomolecule."
        )
        return biomol

    assembly_ids = sorted(list(assembly_gen_category.keys()))
    if assembly_id is None:
        # NOTE: Sorting ensures that the default assembly is the first biological assembly.
        assembly_id = assembly_ids[0]
    elif assembly_id not in assembly_ids:
        raise KeyError(
            f"Biomolecule has no assembly ID `{assembly_id}`. Available assembly IDs: `{assembly_ids}`."
        )

    # Calculate all possible transformations
    transformations = mmcif_parsing.get_transformations(struct_oper_category)

    # Get transformations and apply them to the affected asym IDs
    assembly = None
    for id in assembly_gen_category:
        op_expr = assembly_gen_category[id]["_pdbx_struct_assembly_gen.oper_expression"]
        asym_id_expr = assembly_gen_category[id]["_pdbx_struct_assembly_gen.asym_id_list"]

        # Find the operation expressions for given assembly ID,
        # where we already asserted that the ID is actually present
        if id == assembly_id:
            operations = mmcif_parsing.parse_operation_expression(op_expr)
            asym_ids = asym_id_expr.split(",")
            # Filter affected asym IDs
            sub_assembly = mmcif_parsing.apply_transformations(
                biomol.subset_chains(asym_ids),
                transformations,
                operations,
            )
            # Merge the chains with asym IDs for this operation
            # with chains from other operations
            if assembly is None:
                assembly = sub_assembly
            else:
                assembly += sub_assembly

    assert exists(assembly), f"Assembly with ID `{assembly_id}` not found."
    return assembly


@typecheck
def make_mmcif_features(
    mmcif_object: mmcif_parsing.MmcifObject,
) -> Tuple[FeatureDict, Biomolecule]:
    """Make features from an mmCIF object."""
    input_sequence = "".join(
        mmcif_object.chain_to_seqres[chain_id] for chain_id in mmcif_object.chain_to_seqres
    )
    description = mmcif_object.file_id

    mmcif_feats = {}

    mmcif_feats.update(
        make_sequence_features(
            sequence=input_sequence,
            description=description,
        )
    )

    # As necessary, expand the first bioassembly/model sequence and structure, to obtain a biologically relevant complex (AF3 Supplement, Section 2.1).
    # Reference: https://github.com/biotite-dev/biotite/blob/1045f43f80c77a0dc00865e924442385ce8f83ab/src/biotite/structure/io/pdbx/convert.py#L1441

    assembly = (
        _from_mmcif_object(mmcif_object)
        if "assembly" in description
        else get_assembly(_from_mmcif_object(mmcif_object))
    )

    mmcif_feats["all_atom_positions"] = assembly.atom_positions
    mmcif_feats["all_atom_mask"] = assembly.atom_mask
    mmcif_feats["b_factors"] = assembly.b_factors
    mmcif_feats["chain_index"] = assembly.chain_index
    mmcif_feats["chain_id"] = assembly.chain_id
    mmcif_feats["chemid"] = assembly.chemid
    mmcif_feats["chemtype"] = assembly.chemtype
    mmcif_feats["residue_index"] = assembly.residue_index
    mmcif_feats["restype"] = assembly.restype

    mmcif_feats["bonds"] = mmcif_object.bonds

    mmcif_feats["resolution"] = np.array([mmcif_object.header["resolution"]], dtype=np.float32)

    mmcif_feats["release_date"] = np.array(
        [mmcif_object.header["release_date"].encode("utf-8")], dtype=object
    )

    mmcif_feats["is_distillation"] = np.array(0.0, dtype=np.float32)

    return mmcif_feats, assembly


if __name__ == "__main__":
    filepath = os.path.join("data", "test", "7a4d-assembly1.cif")
    file_id = os.path.splitext(os.path.basename(filepath))[0]

    mmcif_object = mmcif_parsing.parse_mmcif_object(
        filepath=filepath,
        file_id=file_id,
    )
    mmcif_feats, assembly = make_mmcif_features(mmcif_object)
    cropped_assembly, _, _ = assembly.crop(
        contiguous_weight=0.2,
        spatial_weight=0.4,
        spatial_interface_weight=0.4,
        n_res=384,
        chain_1="A",
        chain_2="B",
    )
    mmcif_string = to_mmcif(
        # assembly,
        cropped_assembly,
        file_id=file_id,
        gapless_poly_seq=True,
        insert_alphafold_mmcif_metadata=False,
        # unique_res_atom_names=assembly.unique_res_atom_names,
        unique_res_atom_names=cropped_assembly.unique_res_atom_names,
    )
    with open(os.path.basename(filepath).replace(".cif", "_reconstructed.cif"), "w") as f:
        f.write(mmcif_string)

    print(f"Successfully reconstructed {filepath} after mmCIF featurization.")



import os

import pytest

from alphafold3_pytorch.inputs import PDBDataset
from alphafold3_pytorch.data.weighted_pdb_sampler import WeightedPDBSampler
from alphafold3_pytorch.trainer import pdb_inputs_to_batched_atom_input
from alphafold3_pytorch.utils.utils import exists


def test_template_loading():
    """Test a template-featurized PDBDataset constructed using a WeightedPDBSampler."""
    data_test = os.path.join("data", "test")
    if not os.path.exists(os.path.join("data", "test", "mmcif")):
        pytest.skip("The directory `data/test/mmcif` is not populated yet.")

    interface_mapping_path = os.path.join(data_test, "interface_cluster_mapping.csv")
    chain_mapping_paths = [
        os.path.join(data_test, "ligand_chain_cluster_mapping.csv"),
        os.path.join(
            data_test,
            "nucleic_acid_chain_cluster_mapping.csv",
        ),
        os.path.join(data_test, "peptide_chain_cluster_mapping.csv"),
        os.path.join(data_test, "protein_chain_cluster_mapping.csv"),
    ]

    sampler = WeightedPDBSampler(
        chain_mapping_paths=chain_mapping_paths,
        interface_mapping_path=interface_mapping_path,
        batch_size=64,
    )

    pdb_input = PDBDataset(
        folder=os.path.join("data", "test", "mmcif"),
        sampler=sampler,
        sample_type="default",
        crop_size=128,
        templates_dir=os.path.join("data", "test", "template"),
        training=False,
    )

    batched_atom_input = pdb_inputs_to_batched_atom_input(pdb_input[0], atoms_per_window=27)
    assert exists(batched_atom_input)
